import requests
import getpass
import json

import time
from functools import wraps
from requests.adapters import HTTPAdapter

import logging
import logzero
from logzero import logger


def log(func):
    @wraps(func)
    def wrapped(*args, **kwargs):
        logger.info(f'Calling {func.__name__}()')
        logger.debug(f'kwargs in: {kwargs}')

        start = time.time()
        func_call = func(*args, **kwargs)
        diff = start - time.time()
        logger.debug(f'Call took {diff}s')

        return func_call
    return wrapped


class MMClient:
    """MobilityMaster API Client object.

    All AOS8 API resources are found on the MM https://<MM-IP|FQDN>:4343/api

    Params
    ------
    mm_host: `str`, optional, default: None
        FQDN or IP of the Mobility Master. No leading https://

    username: `str`, optional, default: None
        Username for Mobility Master login

    password: `str`, optional, default: None
        Password for Mobility Master login

    api_version: `str`, optional, default: 1
        The API version used for the calls.

    port: `str`, optional, default: 4343
        The port for the connection

    verify: `str`, optional, default: False
        Same as requests verify, but defaults to False as most MM
        implementations are expected to not use certs. Either a boolean, in
        which case it controls whether we verify the server’s TLS
        certificate, or a string, in which case it must be a path to a CA
        bundle to use.

    timeout: `int`, optional, default: 10
        The timeout for the connection

    Examples
    --------
    **Ex. 1:** Passing in minimum required parameters

    mmc = MMClient(
            mm_host="arubamm.domain.com",
            username="apiuser",
            password="the_password"
            )

    Instance can be created without some or all of above attributes. User
    will be prompted for either one or more `mm_host`, `username`,
    `password`.

    **Ex. 2:** passing in just the host.

    >>> mmc = MMClient(mm_host="arubamm.domain.com")
    MM API username required. Should I use `donkey.kong` to continue [Y/n]?
    MM API password for user `donkey.kong` required:

    **Ex. 3:** Passing in just the username and password

    >>> password = getpass.getpass()
    Password:

    >>> mmc = MMClient(username="apitest", password=password)
    Mobility Master URL or IP required: arubamm.domain.com

    **Ex. 4:** Using a proxy
    To use a proxy specify it with the `proxy` parameter
    """

    def __init__(self, mm_host=None, username=None, password=None, api_version=1, port=4343, verify=False, timeout=10, proxy=str()):
        # Set default logging to error_resp
        logzero.loglevel(logging.ERROR)

        self.mm_host = mm_host
        self.username = username
        self.password = password
        self.api_version = api_version
        self.port = port
        self.timeout = abs(timeout)
        self._access_token = ""

        self.proxy = {}
        if proxy:
            self.proxy = {
                'http': proxy,
                'https': proxy
                }

        self.verify = verify
        if self.verify == False:
            # Disable warnings that come up, as we're not checking the cert
            requests.packages.urllib3.disable_warnings()
            logger.info("Not verifying SSL")

    @log
    def comms(self):
        """User prompt for getting username and/or password, if they haven't been
        passed in with the constructor.
        """
        #
        # If MM URL or IP is not provided, ask for it
        #
        if not self.mm_host:
            self.mm_host = input("Mobility Master URL or IP required: ")

        # If username not provided...
        if not self.username:
            # ... ask for it
            user_input = input("MM API username required. Should I use `{}` to continue [Y/n]?".format(getpass.getuser()))

            # Option for a user if they want to specify a username
            if user_input.lower() == "n":
                self.username = input("API username:\x20")
            # ...any other answer, just use their current username
            else:
                self.username = getpass.getuser()

        # If password not provided, ask for it
        if not self.password:
            # If password is not set, get the username to log into and ask the
            # user for the password for that user
            self.password = getpass.getpass("MM API password for user `{}` required:\x20".format(self.username))

        # Base API URL for requests
        self.mm_base_api_url = f"{self.mm_host}:{self.port}/v{self.api_version}"
        if "https://" not in self.mm_base_api_url:
            self.mm_base_api_url = f"https://{self.mm_base_api_url}"

        # The login credentials dictionary
        self.login_payload = {
            'username': self.username,
            'password': self.password,
        }

        # We require data be returned in JSON format
        # We also pass it in the same format
        self.headers = {
            'Content-Type': 'application/json',
            'Accept' : 'application/json',
        }

        # Configure the session
        self.session = requests.Session()
        self.session.headers.update(self.headers)
        self.session.verify = self.verify
        self.session.timeout = self.timeout
        self.session.proxies = self.proxy
        self.session.mount(self.mm_base_api_url, HTTPAdapter(max_retries=3))

        assert_status_hook = lambda response, *args, **kwargs: response.raise_for_status()
        self.session.hooks["response"] = [assert_status_hook]

        # Finaly login
        self._login()

    @log
    def _params(self, **kwargs):
        '''Parameters configurator for passing into the requests module

        Meant for parameters that get passed with the `params` attribute of
        requests.

        Args:
        -----
        config_path: `str`, default '/md'
            The config path of the MM.

        profile_name: `list`, optional
            A full or partial list of profile names to be searched on. If
            partial include the proper 'filter_oper' otherwise `$eq` will be
            used to match exactly for that name.

        filter_oper: `str`, optional, default '$eq'
            An Aruba API defined filter operator:
                $eq: matches one of the values
                $neq: does not match any of the values
                $gt: matches a value which is greater than the filter
                $gte: matches a value which is greater than or equal to the filter
                $lt: matches a value which is less than the filter
                $lte: matches a value which is less than or equal to the filter
                $in: pattern matches the filter value. E.g., if filter says “ap”,
                “default-ap” and “ap- grp1” will both match
                $nin: pattern does not match the filter. Opposite of $in

        filter: `str`, optional
            A JSON data filter expression for the GET request.
            If defined it will override the `profile_name` and `filter_oper`
            arguments completely.

            Ex.:
            [ {"ap_sys_prof.profile-name" : { "$in" : ["def"] } } ]
            returns all profiles who's names include the 'def' string

            [ {"ht_ssid_prof.profile-name" : { "$eq" : ["default"] } } ]
            returns the profile who's name matches the whole 'default' string

        limit: `str`, optional
            The maximum number of instances of an object that a single GET
            request should return.

        count: `str`, optional
            List of fully qualified parameter name for which count operation
            needs to be performed.
            Ex.: 'int_vlan.int_vlan_ip_helper'

        total: `str`, optional
            The total number of instances of that object existing in system at
            that given configuration node. It should be set to 0 when first
            query is done and the count field specified in the result should be
            put here (optionally) for subsequent queries.

        offset: `str`, optional
            Conveys the number of the entry from which we should start the next
            data set.

        sort: `str`, optional
            The data from the GET request can be sorted based on a single field
            (currently, multi-parameter or nested sorts are not supported).
            There can only be one sort filter per request.
            Works in the form or <oper><key>, where <oper> is either + or - and
            <key> is the fully qualified object to which to sort on

            **Ex.:**
            '-int_vlan.int_vlan_mtu.value' will sort in descending order on the
            value parameter of 'int_vlan.int_vlan_mtu' object

        Returns:
        --------
        params: dict
            The full dict of parameters to be passed with the requets params attribute
        '''
        # Set the default config path if none provided
        params = {
            "config_path": kwargs.get('config_path', '/md'),
            "UIDARUBA": self._access_token
        }

        # If config path was provided with the kwargs, then rewrite the defualt
        if 'config_path' in kwargs:
            params['config_path'] = kwargs['config_path']

        # If `profile_name` provided filter for its exact match in the request
        # and use the `filter_oper` if provided
        if 'profile_name' in kwargs:
            # Send the profile-name in as list instead of just string
            profile_names_list = list()
            if not isinstance(kwargs['profile_name'], list):
                profile_names_list.append(kwargs['profile_name'])
            else: profile_names_list = kwargs['profile_name']

            profile_name_filter = [
                {
                    f"{kwargs['search']}.profile-name": { kwargs.get('filter_oper', '$eq'): profile_names_list },
                }
            ]
            params['filter'] = json.dumps(profile_name_filter)

        # If filter provided it will override whatever was passed in with
        # either `profile_name` or `filter_oper` attributes
        if 'filter' in kwargs:
            if type(kwargs['filter']) is str:
                params['filter'] = kwargs['filter']
            else:
                params['filter'] = json.dumps(kwargs['filter'])

        # Other optional Aruba API defined parameters
        if 'limit' in kwargs:
            params['limit'] = str(kwargs['limit'])
        if 'count' in kwargs:
            params['count'] = str(kwargs['count'])
        if 'total' in kwargs:
            params['total'] = str(kwargs['total'])
        if 'offset' in kwargs:
            params['offset'] = str(wargs['offset'])
        if 'sort' in kwargs:
            params['sort'] = str(kwargs['sort'])
        if 'offset' in kwargs:
            params['offset'] = str(kwargs['offset'])

        logger.debug(f"Returned params: {params}")

        return params

    @log
    def _resource_url(self, resource):
        '''The resource URL formatter

        Will add the preceding '/' if not provided in the resource string

        Args:
        -----
        resource: `str`
            The endpoint resource, ex. 'configuration/object/ap_sys_prof'

        Returns:
        --------
        The full URL string to the requested resource.
        '''
        url = str()

        if resource.startswith("/"):
            url = f'{self.mm_base_api_url}{resource}'
            logger.debug(f"URL to endpoint: {url}")
            return url
        else:
            url = f'{self.mm_base_api_url}/{resource}'
            logger.debug(f"URL to endpoint: {url}")
            return url

    @log
    def _api_call(self, method, url, **kwargs):
        '''The API call handler.

        This method is used by `resource`. kwargs passed in get passed to the
        requests.session instance

        Args:
        -----
        method: `str`
            either `POST` or `GET`

        url: `str`
            URL with the endpoint included

        **kwargs:
        These are passed into the requests and include the `params` and `json`
        attributes which are the exact same ones used by requests.

        Returns:
        --------
        The full response in JSON format including `_global_result` AND
        The error if status string returned is not 0, else `None`.
        '''
        logger.info(f"Method is: {method.upper()}")

        response = getattr(self.session, method.lower())(url, verify=self.verify, **kwargs)
        logger.debug(f"Full URL: {response.url}")

        # If response is wrong, for example if someone passes in the wrong
        # endpoint return https://gitlab.ocado.tech/Net-wifi/wireless-passphrase-change/-/merge_requests/2, None for both values
        try:
            jresp = response.json()
            logger.debug(f"Response JSON: {jresp}")
        except json.decoder.JSONDecodeError as e:
            logger.exception(f'Got a JSONDecodeError exception. Check the defined endpoint is correct\n')
            logger.exception(f"Response text:\n{response.text}")
            logzero.loglevel(logging.ERROR)
            return None, None

        logzero.loglevel(logging.ERROR)

        # Return propper values depending on the type of HTTP request and
        # the response received from it
        if method.lower() == "get":
            return jresp, None
        else:
            if '_global_result' in jresp:
                if jresp["_global_result"]["status"] == 0 or jresp["_global_result"]["status"] == '0':
                    return jresp, None
                else:
                    #logger.debug(f'Error is: {jresp["_global_result"]}')
                    return jresp, jresp["_global_result"]
            else:
                return None, logger.error(f"Config not written: {jresp}")

    @log
    def _login(self):
        '''Login handler to loginto the Mobility Master server.

        Returns:
        --------
        Either the response from the sucesfull login attempt or the error from
        it.
        '''
        logger.info(f"SSL verify (False or cert path): {self.verify}")

        login_url = f'{self.mm_base_api_url}/api/login'

        error_msg = f'Calling _login() with host: {self.mm_host} & username: {self.username}'

        try:
            login_resp, login_resp_err = self._api_call("post", login_url, data=self.login_payload)

        except requests.exceptions.ConnectionError as exc:
            logger.exception(f'Got a {exc.__class__.__name__} exception\n')
            logger.error(error_msg)
            exit(0)
        except requests.RequestException as exc:
            logger.exception(f'Got a {exc.__class__.__name__} exception\n')
            logger.error(error_msg)
            exit(0)
        except requests.HTTPError as exc:
            logger.exception(f'Got a {exc.__class__.__name__} exception\n')
            logger.error(error_msg)
            exit(0)
        except requests.URLRequired as exc:
            logger.exception(f'Got a {exc.__class__.__name__} exception\n')
            logger.error(error_msg)
            exit(0)
        except requests.TooManyRedirects as exc:
            logger.exception(f'Got a {exc.__class__.__name__} exception\n')
            logger.error(error_msg)
            exit(0)
        except requests.ConnectTimeout as exc:
            logger.exception(f'Got a {exc.__class__.__name__} exception\n')
            logger.error(error_msg)
            exit(0)
        except requests.ReadTimeout as exc:
            logger.exception(f'Got a {exc.__class__.__name__} exception\n')
            logger.error(error_msg)
            exit(0)

        # Set the UIDARUBA as the API token
        if not login_resp_err:
            self._access_token = login_resp['_global_result']['UIDARUBA']
            return login_resp
        else:
            logger.error(f"Login failed:\n{login_resp_err}")
            return login_resp_err

    @log
    def _kwargs_modify(self, api_endpoint, data=None, **kwargs):
        '''Modifies the `kwargs` depending if either POST-ing data or GET-ing
        it.

        This is used with resource methods where it adds keys to the `kwargs`
        allready passed in.

        Args:
        -----
        api_endpoint: `str`
            The hardcoded endpoint for the specific resource method.
        data: dict, optional
            If data passed in the HTTP method defaults to POST and the
        **kwargs:
            These are passed into `self._params` and used in requests with the
            params attribute.

        Returns:
        --------
        Modified `kwargs` passed in with the method.
        '''
        logger.debug(f'Endpoint: {api_endpoint}')
        logger.debug(f'Data Payload: {data}')

        kwargs['endpoint'] = api_endpoint

        # If data is passed in, the HTTP method is POST with that data
        if data:
            kwargs['method'] = 'POST'
            kwargs['jpayload'] = data
        else:
            kwargs['method'] = 'GET'
            kwargs['search'] = kwargs['endpoint'].split('/')[-1]

        logger.debug(f'kwargs out: {kwargs}')

        return kwargs

    @log
    def logout(self):
        '''Logs out of the current instance of MM

        Returns:
        --------
        The full response in JSON format including `_global_result`

        The error if status string returned is not 0, else it returns None
        '''
        logout_url = f'{self.mm_base_api_url}/api/logout'

        jresp, jresp_err = self._api_call("get", logout_url)

        return jresp, jresp_err

    @log
    def write_mem(self, config_path=None):
        '''Saves the config at the given `config_path` level.

        Defaults to `/md` if none provided.

        Args:
        -----
        config_path: `str`
            The level at which the config will be saved at.

        Returns:
        --------
        The full response in JSON format including `_global_result`

        The error if status string returned is not 0, else it returns None
        '''
        url = f'{self.mm_base_api_url}/configuration/object/write_memory'
        params = self._params(config_path=config_path)

        jresp, jresp_err = self._api_call("POST", url, params=params)

        return jresp, jresp_err

    @log
    def resource(self, method, endpoint, jpayload=None, **kwargs):
        '''Actiones the HTTP request type defined with the `method` attribute to
        the defined `endpoint`.

        This is the main function of the class, which does all the interfacing
        with the API. It can be called by its own to GET or POST to a resource,
        but the preffered way is to define a resource method below that utilises
        this one for interfacing. The difference between the 2 apporoaches is
        shown in the Examples.

        Args:
        -----
        method: `str`
            Either `GET` or `POST`. Case insensitive.
        endpoint: `str`
            An rendpoint esource path (ex.: "configuration/object/ap_sys_prof").
            Can be with or without the leading `/`.
            The endpoint string is split at `/` and the final list entry used
            as the `search` string used with the `self._params` method.
        jpayload: dict, optional
            JSON formated payload. Same as requests json sent with the body of
            the request. With this passed in, the HTTP method is always POST.
        **kwargs:
            These get passed to the `_params` method, so read what is accepted
            from there.

        Returns:
        --------
        The JSON response and None for error if everything went OK.
        The JSON response and JSON error response in case of the response
        getting an error.

        Examples:
        ---------
        In the examples below the JSON data `apsys_prof` is passed and `aos_obj`
        in an MMClient instance.

        >>> apsys_prof = {
            'bkup_lms_ip': {'bkup-lms-ip': '10.11.11.11'},
            'lms_hold_down_period': {'lms-hold-down-period': 10},
            'lms_ip': {'lms-ip':'10.10.10.11'},
            'lms_preemption': {},
            'profile-name': 'test-01.ap_sys_prof'}

        >>> aos_obj = MMClient()

        **Ex. 1:** Calling a dedicated `/configuration/object/ap_sys_prof` endpoint
        method. (THE PREFFERED WAY)

        a) POST `apsys_prof` data to the '/md/Test_empty' level
        >>> aos_obj.ap_sys_profile(
                data=apsys_prof,
                config_path='/md/Test_empty')

        b) GET the exact ap_sys_profile derfined with
        `profile_name=test-01.ap_sys_prof` at the `/md/Test_empty` level

        >>> aos_obj.ap_sys_profile(
                profile_name='test-01.ap_sys_prof',
                config_path='/md/Test_empty')

        **Ex. 2:** The same as Ex. 1, but with calling the resource method directly.

        a) POST `apsys_prof` data at the '/md/Test_empty' level
        >>> aos_obj.resource(
                'POST',
                endpoint='/configuration/object/ap_sys_prof',
                config_path='/md',
                jpayload=apsys_prof)

        b) GET the exact ap_sys_profile derfined with
        `profile_name=test-01.ap_sys_prof` at the `/md/Test_empty` level

        >>> aos_obj.resource(
                'GET',
                endpoint='configuration/object/ap_sys_prof',
                config_path='/md/Test_empty',
                profile_name='test-01.ap_sys_prof')

        '''
        # Add the 'search' string used with the filter option used by the
        # self._params method.
        # For example this is the last element in the splited endpoint string.
        # For 'configuration/object/ap_sys_prof' this would be 'ap_sys_prof'
        kwargs['search'] = endpoint.split('/')[-1]

        # Get the params from the passed in kwargs
        params = self._params(**kwargs)

        # Build the full URL to the resource
        resource_url = self._resource_url(endpoint)

        # Get the JSON response and error
        jresp, jresp_err = self._api_call(method, resource_url, params=params, json=jpayload)

        return jresp, jresp_err


    '''Below are defined resource methods, which are the ones with a
    hardcoded endpoint object.
    '''
    @log
    def show_command(self, command):
        '''RM to GET a response from a show command.

        If `data` passed method is POST and takes presedence over other
        attributes.

        Args:
        -----
        data: `str`, optional
            JSON formated payload. Same as requests json sent with the body of
            the request. With this passed in, the HTTP method is always POST.
        **kwargs:
            These are passed to `self._kwargs_modify` and `self._params` to
            create a propper request with params.

        Returns:
        --------
        The same as what `self.resource` returns, which is response and
        error or None if no error.
        '''
        logger.debug(f'Data in: {command}')

        kwargs = self._kwargs_modify(
            'configuration/object/ap_sys_prof',
            data,
            **kwargs)

        return self.resource(**kwargs)


    @log
    def ap_sys_profile(self, data=None, **kwargs):
        '''RM to GET or POST to an `ap_sys_prof` endpoint object.

        If `data` passed method is POST and takes presedence over other
        attributes.

        Args:
        -----
        data: `str`, optional
            JSON formated payload. Same as requests json sent with the body of
            the request. With this passed in, the HTTP method is always POST.
        **kwargs:
            These are passed to `self._kwargs_modify` and `self._params` to
            create a propper request with params.

        Returns:
        --------
        The same as what `self.resource` returns, which is response and
        error or None if no error.
        '''
        logger.debug(f'Data in: {data}')

        kwargs = self._kwargs_modify(
            'configuration/object/ap_sys_prof',
            data,
            **kwargs)

        return self.resource(**kwargs)

    @log
    def wlan_ssid_profile(self, data=None, **kwargs):
        '''RM to GET or POST to an `ssid_prof` endpoint object.

        If `data` passed method is POST and takes presedence over other
        attributes.

        Args:
        -----
        data: `str`, optional
            JSON formated payload. Same as requests json sent with the body of
            the request. With this passed in, the HTTP method is always POST.
        **kwargs:
            These are passed to `self._kwargs_modify` and `self._params` to
            create a propper request with params.

        Returns:
        --------
        The same as what `self.resource` returns, which is response and
        error or None if no error.
        '''
        logger.debug(f'Data in: {data}')

        kwargs = self._kwargs_modify(
            'configuration/object/ssid_prof',
            data,
            **kwargs)

        return self.resource(**kwargs)

    @log
    def ap_group(self, data=None, **kwargs):
        '''RM to GET or POST to an `ap_group` endpoint object.

        If `data` passed method is POST and takes presedence over other
        attributes.

        Args:
        -----
        profile_name: ``str``, optional
            If passed in, it will search for that profile
        data: `str`, optional
            JSON formated payload. Same as requests json sent with the body of
            the request. With this passed in, the HTTP method is always POST.
        **kwargs:
            These are passed to `self._kwargs_modify` and `self._params` to
            create a propper request with params.

        Returns:
        --------
        The same as what `self.resource` returns, which is response and
        error or None if no error.
        '''
        logger.debug(f'Data in: {data}')

        kwargs = self._kwargs_modify(
            'configuration/object/ap_group',
            data,
            **kwargs)

        return self.resource(**kwargs)

    @log
    def virtual_ap(self, data=None, **kwargs):
        '''RM to GET or POST to an `virtual_ap` endpoint object.

        If `data` passed method is POST and takes presedence over other
        attributes.

        Args:
        -----
        profile_name: ``str``, optional
            If passed in, it will search for that profile
        data: `str`, optional
            JSON formated payload. Same as requests json sent with the body of
            the request. With this passed in, the HTTP method is always POST.
        **kwargs:
            These are passed to `self._kwargs_modify` and `self._params` to
            create a propper request with params.

        Returns:
        --------
        The same as what `self.resource` returns, which is response and
        error or None if no error.
        '''
        logger.debug(f'Data in: {data}')

        kwargs = self._kwargs_modify(
            'configuration/object/virtual_ap',
            data,
            **kwargs)

        return self.resource(**kwargs)

    @log
    def ap_sys_prof(self, data=None, **kwargs):
        '''RM to GET or POST to an `ap_sys_prof` endpoint object.

        If `data` passed method is POST and takes presedence over other
        attributes.

        Args:
        -----
        profile_name: ``str``, optional
            If passed in, it will search for that profile
        data: `str`, optional
            JSON formated payload. Same as requests json sent with the body of
            the request. With this passed in, the HTTP method is always POST.
        **kwargs:
            These are passed to `self._kwargs_modify` and `self._params` to
            create a propper request with params.

        Returns:
        --------
        The same as what `self.resource` returns, which is response and
        error or None if no error.
        '''
        logger.debug(f'Data in: {data}')

        kwargs = self._kwargs_modify(
            'configuration/object/ap_sys_prof',
            data,
            **kwargs)

        return self.resource(**kwargs)

    @log
    def reg_domain_prof(self, data=None, **kwargs):
        '''RM to GET or POST to an `reg_domain_prof` endpoint object.

        If `data` passed method is POST and takes presedence over other
        attributes.

        Args:
        -----
        profile_name: ``str``, optional
            If passed in, it will search for that profile
        data: `str`, optional
            JSON formated payload. Same as requests json sent with the body of
            the request. With this passed in, the HTTP method is always POST.
        **kwargs:
            These are passed to `self._kwargs_modify` and `self._params` to
            create a propper request with params.

        Returns:
        --------
        The same as what `self.resource` returns, which is response and
        error or None if no error.
        '''
        logger.debug(f'Data in: {data}')

        kwargs = self._kwargs_modify(
            'configuration/object/reg_domain_prof',
            data,
            **kwargs)

        return self.resource(**kwargs)

    @log
    def dot11k_prof(self, data=None, **kwargs):
        '''RM to GET or POST to an `dot11k_prof` endpoint object.

        If `data` passed method is POST and takes presedence over other
        attributes.

        Args:
        -----
        profile_name: ``str``, optional
            If passed in, it will search for that profile
        data: `str`, optional
            JSON formated payload. Same as requests json sent with the body of
            the request. With this passed in, the HTTP method is always POST.
        **kwargs:
            These are passed to `self._kwargs_modify` and `self._params` to
            create a propper request with params.

        Returns:
        --------
        The same as what `self.resource` returns, which is response and
        error or None if no error.
        '''
        logger.debug(f'Data in: {data}')

        kwargs = self._kwargs_modify(
            'configuration/object/dot11k_prof',
            data,
            **kwargs)

        return self.resource(**kwargs)

    @log
    def dot11r_prof(self, data=None, **kwargs):
        '''RM to GET or POST to an `dot11r_prof` endpoint object.

        If `data` passed method is POST and takes presedence over other
        attributes.

        Args:
        -----
        profile_name: ``str``, optional
            If passed in, it will search for that profile
        data: `str`, optional
            JSON formated payload. Same as requests json sent with the body of
            the request. With this passed in, the HTTP method is always POST.
        **kwargs:
            These are passed to `self._kwargs_modify` and `self._params` to
            create a propper request with params.

        Returns:
        --------
        The same as what `self.resource` returns, which is response and
        error or None if no error.
        '''
        logger.debug(f'Data in: {data}')

        kwargs = self._kwargs_modify(
            'configuration/object/dot11r_prof',
            data,
            **kwargs)

        return self.resource(**kwargs)

    @log
    def ap_a_radio_prof(self, data=None, **kwargs):
        '''RM to GET or POST to an `ap_a_radio_prof` endpoint object.

        If `data` passed method is POST and takes presedence over other
        attributes.

        Args:
        -----
        profile_name: ``str``, optional
            If passed in, it will search for that profile
        data: `str`, optional
            JSON formated payload. Same as requests json sent with the body of
            the request. With this passed in, the HTTP method is always POST.
        **kwargs:
            These are passed to `self._kwargs_modify` and `self._params` to
            create a propper request with params.

        Returns:
        --------
        The same as what `self.resource` returns, which is response and
        error or None if no error.
        '''
        logger.debug(f'Data in: {data}')

        kwargs = self._kwargs_modify(
            'configuration/object/ap_a_radio_prof',
            data,
            **kwargs)

        return self.resource(**kwargs)

    @log
    def ht_radio_prof(self, data=None, **kwargs):
        '''RM to GET or POST to an `ht_radio_prof` endpoint object.

        If `data` passed method is POST and takes presedence over other
        attributes.

        Args:
        -----
        data: `str`, optional
            JSON formated payload. Same as requests json sent with the body of
            the request. With this passed in, the HTTP method is always POST.
        **kwargs:
            These are passed to `self._kwargs_modify` and `self._params` to
            create a propper request with params.

        Returns:
        --------
        The same as what `self.resource` returns, which is response and
        error or None if no error.
        '''
        logger.debug(f'Data in: {data}')

        kwargs = self._kwargs_modify(
            'configuration/object/ht_radio_prof',
            data,
            **kwargs)

        return self.resource(**kwargs)

    @log
    def ht_ssid_prof(self, data=None, **kwargs):
        '''RM to GET or POST to an `ht_ssid_prof` endpoint object.

        If `data` passed method is POST and takes presedence over other
        attributes.

        Args:
        -----
        data: `str`, optional
            JSON formated payload. Same as requests json sent with the body of
            the request. With this passed in, the HTTP method is always POST.
        **kwargs:
            These are passed to `self._kwargs_modify` and `self._params` to
            create a propper request with params.

        Returns:
        --------
        The same as what `self.resource` returns, which is response and
        error or None if no error.
        '''
        logger.debug(f'Data in: {data}')

        kwargs = self._kwargs_modify(
            'configuration/object/ht_ssid_prof',
            data,
            **kwargs)

        return self.resource(**kwargs)


    @log
    def rrm_ie_prof(self, data=None, **kwargs):
        '''RM to GET or POST to an `rrm_ie_prof` endpoint object.

        If `data` passed method is POST and takes presedence over other
        attributes.

        Args:
        -----
        data: `str`, optional
            JSON formated payload. Same as requests json sent with the body of
            the request. With this passed in, the HTTP method is always POST.
        **kwargs:
            These are passed to `self._kwargs_modify` and `self._params` to
            create a propper request with params.

        Returns:
        --------
        The same as what `self.resource` returns, which is response and
        error or None if no error.
        '''
        logger.debug(f'Data in: {data}')

        kwargs = self._kwargs_modify(
            'configuration/object/rrm_ie_prof',
            data,
            **kwargs)

        return self.resource(**kwargs)

    @log
    def node_hierarchy(self, data=None, **kwargs):
        '''RM to GET or POST to an `node_hierarchy` endpoint object.

        If `data` passed method is POST and takes presedence over other
        attributes.

        Args:
        -----
        data: `str`, optional
            JSON formated payload. Same as requests json sent with the body of
            the request. With this passed in, the HTTP method is always POST.
        **kwargs:
            These are passed to `self._kwargs_modify` and `self._params` to
            create a propper request with params.

        Returns:
        --------
        The same as what `self.resource` returns, which is response and
        error or None if no error.
        '''
        logger.debug(f'Data in: {data}')

        kwargs = self._kwargs_modify(
            'configuration/object/node_hierarchy',
            data,
            **kwargs)

        return self.resource(**kwargs)

    @log
    def add_configuration_device(self, data, **kwargs):
        '''RM to GET or POST to an `add_configuration_device` endpoint object.

        If `data` passed method is POST and takes presedence over other
        attributes.

        Args:
        -----
        data: `str`
            JSON formated payload. Same as requests json sent with the body of
            the request. With this passed in, the HTTP method is always POST.
        **kwargs:
            These are passed to `self._kwargs_modify` and `self._params` to
            create a propper request with params.

        Returns:
        --------
        The same as what `self.resource` returns, which is response and
        error or None if no error.
        '''
        logger.debug(f'Data in: {data}')

        kwargs = self._kwargs_modify(
            'configuration/object/add_configuration_device',
            data,
            **kwargs)

        return self.resource(**kwargs)

    @log
    def netdst(self, data=None, **kwargs):
        '''RM to GET or POST to an `netdst` endpoint object.

        If `data` passed method is POST and takes presedence over other
        attributes.

        Args:
        -----
        data: `str`, optional
            JSON formated payload. Same as requests json sent with the body of
            the request. With this passed in, the HTTP method is always POST.
        **kwargs:
            These are passed to `self._kwargs_modify` and `self._params` to
            create a propper request with params.

        Returns:
        --------
        The same as what `self.resource` returns, which is response and
        error or None if no error.
        '''
        logger.debug(f'Data in: {data}')

        kwargs = self._kwargs_modify(
            'configuration/object/netdst',
            data,
            **kwargs)

        return self.resource(**kwargs)

    @log
    def netsvc(self, data=None, **kwargs):
        '''RM to GET or POST to an `netsvc` endpoint object.

        If `data` passed method is POST and takes presedence over other
        attributes.

        Args:
        -----
        data: `str`, optional
            JSON formated payload. Same as requests json sent with the body of
            the request. With this passed in, the HTTP method is always POST.
        **kwargs:
            These are passed to `self._kwargs_modify` and `self._params` to
            create a propper request with params.

        Returns:
        --------
        The same as what `self.resource` returns, which is response and
        error or None if no error.
        '''
        logger.debug(f'Data in: {data}')

        kwargs = self._kwargs_modify(
            'configuration/object/netsvc',
            data,
            **kwargs)

        return self.resource(**kwargs)

    @log
    def acl_sess(self, data=None, **kwargs):
        '''RM to GET or POST to an `acl_sess` endpoint object.

        If `data` passed method is POST and takes presedence over other
        attributes.

        Args:
        -----
        data: `str`, optional
            JSON formated payload. Same as requests json sent with the body of
            the request. With this passed in, the HTTP method is always POST.
        **kwargs:
            These are passed to `self._kwargs_modify` and `self._params` to
            create a propper request with params.

        Returns:
        --------
        The same as what `self.resource` returns, which is response and
        error or None if no error.
        '''
        logger.debug(f'Data in: {data}')

        kwargs = self._kwargs_modify(
            'configuration/object/acl_sess',
            data,
            **kwargs)

        return self.resource(**kwargs)

    @log
    def aaa_prof(self, data=None, **kwargs):
        '''RM to GET or POST to an `aaa_prof` endpoint object.

        If `data` passed method is POST and takes presedence over other
        attributes.

        Args:
        -----
        data: `str`, optional
            JSON formated payload. Same as requests json sent with the body of
            the request. With this passed in, the HTTP method is always POST.
        **kwargs:
            These are passed to `self._kwargs_modify` and `self._params` to
            create a propper request with params.

        Returns:
        --------
        The same as what `self.resource` returns, which is response and
        error or None if no error.
        '''
        logger.debug(f'Data in: {data}')

        kwargs = self._kwargs_modify(
            'configuration/object/aaa_prof',
            data,
            **kwargs)

        return self.resource(**kwargs)

    @log
    def rad_server(self, data=None, **kwargs):
        '''RM to GET or POST to an `rad_server` endpoint object.

        If `data` passed method is POST and takes presedence over other
        attributes.

        Args:
        -----
        data: `str`, optional
            JSON formated payload. Same as requests json sent with the body of
            the request. With this passed in, the HTTP method is always POST.
        **kwargs:
            These are passed to `self._kwargs_modify` and `self._params` to
            create a propper request with params.

        Returns:
        --------
        The same as what `self.resource` returns, which is response and
        error or None if no error.
        '''
        logger.debug(f'Data in: {data}')

        kwargs = self._kwargs_modify(
            'configuration/object/rad_server',
            data,
            **kwargs)

        return self.resource(**kwargs)

    @log
    def server_group_prof(self, data=None, **kwargs):
        '''RM to GET or POST to an `server_group_prof` endpoint object.

        If `data` passed method is POST and takes presedence over other
        attributes.

        Args:
        -----
        data: `str`, optional
            JSON formated payload. Same as requests json sent with the body of
            the request. With this passed in, the HTTP method is always POST.
        **kwargs:
            These are passed to `self._kwargs_modify` and `self._params` to
            create a propper request with params.

        Returns:
        --------
        The same as what `self.resource` returns, which is response and
        error or None if no error.
        '''
        logger.debug(f'Data in: {data}')

        kwargs = self._kwargs_modify(
            'configuration/object/server_group_prof',
            data,
            **kwargs)

        return self.resource(**kwargs)

    @log
    def role(self, data=None, **kwargs):
        '''RM to GET or POST to an `role` endpoint object.

        If `data` passed method is POST and takes presedence over other
        attributes.

        Args:
        -----
        data: `str`, optional
            JSON formated payload. Same as requests json sent with the body of
            the request. With this passed in, the HTTP method is always POST.
        **kwargs:
            These are passed to `self._kwargs_modify` and `self._params` to
            create a propper request with params.

        Returns:
        --------
        The same as what `self.resource` returns, which is response and
        error or None if no error.
        '''
        logger.debug(f'Data in: {data}')

        kwargs = self._kwargs_modify(
            'configuration/object/role',
            data,
            **kwargs)

        return self.resource(**kwargs)

    @log
    def acl_sess(self, data=None, **kwargs):
        '''RM to GET or POST to an `acl_sess` endpoint object.

        If `data` passed method is POST and takes presedence over other
        attributes.

        Args:
        -----
        data: `str`, optional
            JSON formated payload. Same as requests json sent with the body of
            the request. With this passed in, the HTTP method is always POST.
        **kwargs:
            These are passed to `self._kwargs_modify` and `self._params` to
            create a propper request with params.

        Returns:
        --------
        The same as what `self.resource` returns, which is response and
        error or None if no error.
        '''
        logger.debug(f'Data in: {data}')

        kwargs = self._kwargs_modify(
            'configuration/object/acl_sess',
            data,
            **kwargs)

        return self.resource(**kwargs)



import requests
import xmltodict
import socket
import getpass

import logging
import logzero
from logzero import logger

class OnlyOneInstance(type):
    """Metaclass that allows only one class instance to be created
    """
    _instances = {}

    def __call__(cls, *args, **kwargs):
        if cls not in cls._instances:
            cls._instances[cls] = super(OnlyOneInstance, cls).__call__(*args, **kwargs)
        return cls._instances[cls]


# Class to connect to the Aruba Airwave management system
class AirWave(metaclass=OnlyOneInstance):
    """All Aruba Mobility Master API URIs are found on https://mm-host/api/
    and are accessible if logged in.

    Only the ``aw_url`` is mandatory and username and password will be asked for
    when calling the ``comms()`` method if not provided.

    Parameters
    ----------
    aw_url: `str`
        The URL of the AirWave instance

    aw_username: `str`, optional
        A username for loggin in to AirWave. Although optional it is required
        and will be prompetd for if not passed in.

    aw_password: `str`, optional
        Password for the provided username. Although optional it is required
        and will be prompetd for if not passed in.

    proxy: `str`, optionall
        A proxy to be used. This is used for HTTP and HTTPS connections

    verify: bool or string, optional, deafult false
        Specify how SSL verification will be handled. Handling is done same way
        as with requests, so either:
        - leave blank or use False (bool) to disable OR
        - specify path to a cert file to enable

    timeout: `int` optional, default 30s
        The timeout for the connection.

    Examples:
    ---------
    **Ex. 1:** Importing the AirWave module

    >>> from arubafi import AirWave
    >>> aw = AirWave(
            aw_url="my.airwave.com",
            aw_username="a-username",
            aw_password="a-pass")
    >>> aw.comms()

    **Ex. 2:** Not passing in username or password
    In this case you are asked for username and password or you can provide one
    or both when creating a new instance.

    from arubafi import AirWave
    aw = AirWave(aw_url="my.airwave.com")

    Airwave username required. Should I use `your.username` to continue [Y/n]?
    AirWave Password required:
    <arubatools.airwave.AirWave at 0x112a92dd0>
    """
    def __init__(self, aw_url, aw_username=str(), aw_password=str(), proxy=str(), verify=False, timeout=30):
        self.aw_url = str(aw_url)
        self.aw_username = str(aw_username)
        self.aw_password = str(aw_password)
        self.verify = verify
        self.timeout = timeout

        self.proxy = {}
        if proxy:
            self.proxy = {
                'http': proxy,
                'https': proxy
                }

        # URLs used with AirWave to get client data and to login
        if "https://" not in self.aw_url:
            self.aw_url = f"https://{self.aw_url}"
        self.login_url = self.aw_url + '/LOGIN'

        # Set logging to ERROR to not display anything by default
        logzero.loglevel(logging.ERROR)

    def comms(self):
        """User prompt for getting username and/or password, if they haven't been
        passed in with the constructor.
        """
        logger.info('Calling comms()')

        # If user is not set...
        if not self.aw_username:
            # ...ask for it and get the password regardless of the answer
            system_user = getpass.getuser()
            user_input = input(f"Airwave username required. Should `{system_user}` be used [Y/n]?")

            # Option for a user if they want to specify a username
            if user_input.lower() == "n":
                self.aw_username = input("AirWave username:\x20")
            # ...any other answer, just use their system username
            else:
                self.aw_username = system_user

            # Ask for the password for the user regardless of it being set
            # Can't rememeber the exact reason or this TBH
            self.aw_password = getpass.getpass("AirWave Password required:\x20")

        if not self.aw_password:
            # If password is not set, get the login username and ask
            # for the password for that user
            self.aw_password = getpass.getpass(f"AirWave Password for user `{self.aw_username}` required:\x20")

        # The login credentials payload to send in the request
        self.login_payload = {
            'destination': '/',
            'credential_0': self.aw_username,
            'credential_1': self.aw_password
        }

        # Create a session object
        self.session = requests.Session()
        self.session.proxies = self.proxy
        self.session.verify = self.verify
        self.session.timeout = self.timeout

        if self.verify == False:
            # Disable warnings that come up, as we're not checking the cert
            requests.packages.urllib3.disable_warnings()
            logger.info("Not verifying SSL")

        error_msg = f'Logging into {self.login_url} with username: {self.aw_username}'

        try:
            login_resp = self.session.post(
                self.login_url,
                data=self.login_payload,
                )
            # Raise an exception if the status code != 2xx
            login_resp.raise_for_status()
            # If there is no raise response you're authenticated!

        except requests.exceptions.ConnectionError as exc:
            logger.exception(f'Got a {exc.__class__.__name__} exception\n')
            logger.error(error_msg)
            exit(0)
        except requests.RequestException as exc:
            logger.exception(f'Got a {exc.__class__.__name__} exception\n')
            logger.error(error_msg)
            exit(0)
        except requests.HTTPError as exc:
            logger.exception(f'Got a {exc.__class__.__name__} exception\n')
            logger.error(error_msg)
            exit(0)
        except requests.URLRequired as exc:
            logger.exception(f'Got a {exc.__class__.__name__} exception\n')
            logger.error(error_msg)
            exit(0)
        except requests.TooManyRedirects as exc:
            logger.exception(f'Got a {exc.__class__.__name__} exception\n')
            logger.error(error_msg)
            exit(0)
        except requests.ConnectTimeout as exc:
            logger.exception(f'Got a {exc.__class__.__name__} exception\n')
            logger.error(error_msg)
            exit(0)
        except requests.ReadTimeout as exc:
            logger.exception(f'Got a {exc.__class__.__name__} exception\n')
            logger.error(error_msg)
            exit(0)


    #
    ##
    ### Health Methods
    ### Take care of health checks and such
    ##
    #
    def close(self):
        """Function that just closes the AirWave session
        """
        logger.info('Calling close()')
        self.session.close()

    def _dns_ptr_check(self, addr):
        """Returns either `FQDN` of an IP or `None` which signifies missing PTR
        """
        logger.info(f'Calling _dns_ptr_check({addr})')
        try:
            get_response = socket.gethostbyaddr(addr)
            logging.debug(get_response)

            # Fix for the Apline Python docker image
            if get_response[0] == addr:
                return None, None, None

            return get_response

        except socket.herror:
            return None, None, None


    #
    ##
    ### Inventory Methods
    ### These just fill up the DB dicts from AirWave's output
    ### Do NOT call these methods directly, but use associated ``get_`` methods
    ### instead.
    ##
    #
    def _full_raw_airwave_inventory(self, return_in_dict=True):
        """Returns the WHOLE inventory in a dict (APs, IAPs, controllers and VCs) as returned
        by passing in ~/ap_list.xml without any parameters
        """
        logger.info('Calling _full_raw_airwave_inventory()')
        if not hasattr(self, '_inventory') or not self._inventory:

            ap_list_url = self.aw_url + '/ap_list.xml'

            response = self.session.get(ap_list_url, timeout=30)
            # Raise an exception if the status code != 2xx
            response.raise_for_status()

            # Convert the XML client_detail_resp into a dictionary
            self._inventory = xmltodict.parse(response.content)

            if return_in_dict:
                return self._inventory
            else:
                return response

    def _create_inventory_dbs(self):
        """Creates inventory dictionaries (DBs), split into dicts for APs/IAPs,
        controllers and virtual cotrollers

        GET the DB by calling associated ``get_*`` methods

        Creates:
        - controller_id to controller_fqdn mapping
        - VC_id to VC_fqdn mapping
        - controller_id to ap_name mapping
        - ap_name to controller_id mapping
        - controller_id to controller_fqdn mapping
        - controller_id to controller_name mapping
        - DB for controllers with no FQDN (missing PTR record)
        """
        logger.info('Calling _create_inventory_dbs()')

        # controller_id to controller_fqdn mapping
        self._controllers_db = dict()
        # VC_id to VC_fqdn mapping
        self._iapvc_db = dict()
        # controller_id to ap_name mapping
        self._contrlollerid_to_ap_db = dict()
        # ap_name to controller_id mapping
        self._apname_to_controllerid_db = dict()
        # All standalone APs with no controllermanagement
        self._controllerless_ap_db = dict()
        # DB for controllers with no FQDN (missing PTR record)
        # controller_id to controller_name mapping
        self._no_ptr_controllers_db = dict()
        # A DB for every item with an ID and another dict for various values
        self._all_items_db = dict()


        # get the full AMP inventory dict
        amp_inventory = self._full_raw_airwave_inventory()
        logger.debug(amp_inventory)

        # Iterate through all items and asign them to their respective DB dict
        for item in amp_inventory["amp:amp_ap_list"]["ap"]:

            item_data = self._all_items_db.setdefault(item["@id"], {})

            item_data["lan_ip"] = item.get("lan_ip")
            item_data["lan_mac"] = item.get("lan_mac")
            item_data["name"] = item.get("name")
            item_data["serial_number"] = item.get("serial_number")
            item_data["device_category"] = item.get("device_category")
            item_data["controller_id"] = item.get("controller_id")
            item_data["fqdn"] = item.get("fqdn")
            item_data["manufacturer"] = item.get("mfgr")
            item_data["model"] = item.get("model", {}).get("#text")
            item_data["is_rap"] = item.get("is_remote_ap")



            # Add @id and controller_fqdn as key/value to the _controllers_db
            # if controller is a "normal" controller
            if "controller" in item["device_category"] and "Instant Virtual Controller" not in item['model']['#text']:

                #try getting controllers FQDN, but if key does not exists...
                try:
                    self._controllers_db[item['@id']] = item['fqdn']
                    #print('FQDN', item['@id'], item['fqdn'])
                # ... Try finding it's FQDN via its IP address
                except KeyError:
                    controller_ip = item["lan_ip"]
                    controller_fqdn = self._dns_ptr_check(controller_ip)
                    #print(controller_fqdn)
                    if controller_fqdn[0]:
                        self._controllers_db[item['@id']] = controller_fqdn[0]
                        #print(controller_fqdn[0])1
                    else:
                        self._no_ptr_controllers_db[controller_ip] = item["name"]
                        #print(item["name"])

                    #print("WARNING: no FQDN for {}".format(item["name"]))

            # Add @id and iapvc_fqdn as key/value to the _iapvc_db
            # if the controller is a virtual IAP controller
            elif "controller" in item['device_category'] and "Instant Virtual Controller" in item['model']['#text']:
                self._iapvc_db[item['@id']] = item['fqdn']

            # if item is AP & managed by a controller...
            elif "thin_ap" in item['device_category'] and "controller_id" in item:

                # if the controller_id is not in the dict
                # put it there with a default empty list
                k = item['controller_id']
                self._contrlollerid_to_ap_db.setdefault(item_data["controller_id"], [])
                # Put the AP in the list next to the controller_id
                self._contrlollerid_to_ap_db[k].append(item_data["name"])

                # Put the ap_name with its controller_id into the dict
                self._apname_to_controllerid_db[item_data["name"]] = item_data["controller_id"]


            # if item is AP and is not managed by a controller
            # i.e. it doesn't have a `controller_id`
            elif not "controller_id" in item and not "controller" in item_data["device_category"]:
                # Put it into the controllerless AP database with
                # its own id and its name (AP name)
                self._controllerless_ap_db[item['@id']] = item['name']

    def _controller_inventory(self):
        """Fills up the `self._controllers_db` dict with
        controller ID to controller FQDN key value mappings

        Call `get_controller_inventory()` to get the output of this.
        """
        logger.info('Calling _controller_inventory()')

        if not hasattr(self, '_controllers_db') or not self._controllers_db:
            self._create_inventory_dbs()

        logger.debug(f'Data returned\n{self._controllers_db}')
        logzero.loglevel(logging.ERROR)

        return self._controllers_db

    def _no_ptr_controller_inventory(self):
        """Fills up the `self._no_ptr_controllers_db` dict with
        controller IP to controller NAME key value mappings

        Call `get_no_ptr_controller_inventory()` to get the output of this.
        """
        logger.info('Calling _no_ptr_controller_inventory()')

        if not hasattr(self, '_no_ptr_controllers_db') or not self._no_ptr_controllers_db:
            self._create_inventory_dbs()

        logger.debug(f'Data returned\n{self._no_ptr_controllers_db}')
        logzero.loglevel(logging.ERROR)

        return self._no_ptr_controllers_db

    def _iapvc_inventory(self):
        """Fills up the `self._iapvc_db` dict with
        controller_id to controller_FQDN key value mappings

        Call `get_iapvc_inventory()` to get the output of this.
        """
        logger.info('Calling _iapvc_inventory()')

        if not hasattr(self, '_iapvc_db') or not self._iapvc_db:
            self._create_inventory_dbs()

        logger.debug(f'Data returned\n{self._iapvc_db}')
        logzero.loglevel(logging.ERROR)

        return self._iapvc_db

    def _controllerid_to_ap_inventory(self):
        """Fills up the `self._contrlollerid_to_ap_db` dict with
        AP_NAME to controller's @id key value mappings.

        Call `get_controllerid_to_ap_inventory()` to get the output of this.
        """
        logger.info('Calling _controllerid_to_ap_inventory()')

        if not hasattr(self, '_contrlollerid_to_ap_db') or not self._contrlollerid_to_ap_db:
            self._create_inventory_dbs()

        logger.debug(f'Data returned\n{self._contrlollerid_to_ap_db}')
        logzero.loglevel(logging.ERROR)

        return self._contrlollerid_to_ap_db

    def _apname_to_controllerid_inventory(self):
        """Fills up the `self._apname_to_controllerid_db` dict with
        AP_NAME to controller_ID key value mappings

        Call `get_apname_to_controllerid_inventory()` to get the output of this.
        """
        logger.info('Calling _apname_to_controllerid_inventory()')

        if not hasattr(self, '_apname_to_controllerid_db') or not self._apname_to_controllerid_db:
            self._create_inventory_dbs()

        logger.debug(f'Data returned\n{self._apname_to_controllerid_db}')
        logzero.loglevel(logging.ERROR)

        return self._apname_to_controllerid_db

    def _controllerless_ap_inventory(self):
        """Fills up the  `self._controllerless_ap_db`

        Call `get_ap_inventory()` to get the output of this.
        """
        logger.info('Calling _controllerless_ap_inventory()')

        if not hasattr(self, '_controllerless_ap_db') or not self._controllerless_ap_db:
            self._create_inventory_dbs()

        logger.debug(f'Data returned\n{self._controllerless_ap_db}')
        logzero.loglevel(logging.ERROR)

        return self._controllerless_ap_db

    def _all_items_inventory(self):
        """Fills up the `self._all_items_db`

        Call `get_all_items_inventory()` to get the output of this.
        """
        logger.info('Calling _all_items_inventory()')

        if not hasattr(self, '_all_items_db') or not self._all_items_db:
            self._create_inventory_dbs()

        logger.debug(f'Data returned\n{self._all_items_db}')
        logzero.loglevel(logging.ERROR)

        return self._all_items_db


    #
    ##
    ### The GET inventory methods
    ### These return a full DB of elements specified within them
    ##
    #
    def get_controller_inventory(self):
        """Returns a dict with controller's @id as key and FQDN as value
        as returned by accessing the AirWave/ap_list.xml without any parameters.

        Example output
        --------------
        controllers_db = {
            '1064': 'loop0-cont01.blah.com',
            '1069': 'loop0-cont02.blah.com',
            '1072': 'loop0-cont03.blah.com',
            '1090': 'loop0-cont04.blah.com',
            '1116': 'loop0-cont05.blah.com'
        }
        """
        logger.info('Calling get_controller_inventory()')
        return self._controller_inventory()

    def get_no_ptr_controller_inventory(self):
        """Returns a dict with controller's @id as key and NAME
        as value as returned by accessing the AirWave/ap_list.xml without any parameters.

        Example output
        --------------
        {'10.238.3.4': 'aa-wi0',
         '10.254.44.19': 'bb-wi0',
         '10.254.44.21': 'cc-wi1'}
        """
        logger.info('Calling get_no_ptr_controller_inventory()')
        return self._no_ptr_controller_inventory()

    def get_iapvc_inventory(self):
        """Returns a dict with controller's @id as key and NAME
        as value as returned by accessing the AirWave/ap_list.xml without any parameters.

        Example output:
        --------------
        {
            '1627': 'th-iapvc0.blah.com'
        }
        """
        logger.info('Calling get_iapvc_inventory()')
        return self._iapvc_inventory()

    def get_controllerid_to_ap_inventory(self):
        """Returns a dict with controller @id as key and its list of associated APS
        as returned by accessing the AirWave/ap_list.xml without any parameters.

        Example output
        --------------
        {
            '1865': ['ap501', 'ap502', 'ap503'],
            '367': ['ap164','ap165','ap167','ap161','ap162','ap163']
        }
        """
        logger.info('Calling get_controllerid_to_ap_inventory()')
        return self._controllerid_to_ap_inventory()

    def get_apname_to_controllerid_inventory(self):
        """Returns a dict with AP `name` as key and its `controller @id` as value
        as returned by accessing the AirWave/ap_list.xml without any parameters.

        DB contains ALL APs (CAP, RAP, IAP, or other)

        Example output
        --------------
        {
            'ap160.zv': '1901',
            'ap120': '1703',
            'ap130': '1901',
            'ap153': '1901',
        }
        """
        logger.info('Calling get_apname_to_controllerid_inventory()')
        return self._apname_to_controllerid_inventory()

    def get_controllerless_ap_inventory(self):
        """Returns the dict of controlerless APs

        key is AP's @id
        value is the AP's name

        Example output
        --------------
        {
            '1106': 'ap12',
            '1109': 'ap13',
            '1394': 'ap31',
            '1399': 'ap183',
            '527': 'ap33',
            '808': 'ap17',
            '948': 'ap20'
        }
        """
        logger.info('Calling get_controllerless_ap_inventory()')
        return self._controllerless_ap_inventory()

    def get_all_items_inventory(self):
        """Returns the inventory dict of ALL items from AW within it

        Example output
        --------------
        {
        '101': {
            'lan_ip': '10.22.50.106',
            'lan_mac': '00:01:86:C2:dd:aa',
            'name': 'ap54',
            'serial_number': '345309345',
            'device_category': 'thin_ap',
            'controller_id': '17',
            'fqdn': 'ap54',
            'manufacturer': 'Aruba',
            'model': 'AP 70',
            'is_rap': 'false'
            },

        '1687': {
            'lan_ip': '10.254.104.21',
            'lan_mac': '00:4C:f3:35:C2:aC',
            'name': 'wi0',
            'serial_number': '345345345',
            'device_category': 'controller',
            'controller_id': None,
            'fqdn': 'wi0-loop.blah.com',
            'manufacturer': 'Aruba',
            'model': '7010',
            'is_rap': None
            },
        ...
        }
        """
        logger.info('Calling get_all_items_inventory()')
        return self._all_items_inventory()


    #
    ##
    ### GET specific elements methods
    ### These return a subset of elements from the DB dicts produced by
    ### the _*() methods of elements specified within them
    ##
    #
    def get_users_ap_info(self, users_mac):
        """
        Gets the AP to which the user is connected to, by passing
        in the `user_mac`

        Example output:

        {
            'ap_id': '1868',
            'name': 'ap502',
            'radio': 'a',
            'essid': 'blah',
            'vlan': '610',
            'client_aw_url': 'https://test.airwave.com/client_detail.xml?mac=33%3Add%34ff%3Aaa%3Abb%3A11',
            'controller_id': '1865',
            'client_count': '17',
            'firmware': '6.5.4.9',
            'ap_fqdn': 'ap502.blah.com',
            'lan_ip': '10.22.72.12',
            'lan_mac': '20:A6:be:C5:3f:d2',
            'model': 'AP 305',
            'operating_mode': 'ap',
            'serial_number': 'FGBTG345345',
            'ap_aw_url': 'https://test.airwave.com/ap_list.xml?id=1868',
        }
        """
        logger.info('Calling get_users_ap_info()')

        # URLs used with Airvawe to get client and client's AP data
        client_detail_url = self.aw_url + '/client_detail.xml'
        ap_list_url = self.aw_url + '/ap_list.xml'

        self.users_ap_info = dict()
        self.user_mac = {'mac': users_mac.upper()}

        # Make a call to AW with users MAC address
        client_detail_resp = self.session.get(client_detail_url, params=self.user_mac, timeout=30)
        logger.debug(client_detail_resp.request.url)
        logger.debug(client_detail_resp.content)
        client_detail_resp.raise_for_status()

        # Convert the XML client_detail_resp into a dictionary
        amp_client_detail = xmltodict.parse(client_detail_resp.content)
        logger.debug(amp_client_detail['amp:amp_client_detail'])

        # Populate the users_ap_info database with information but check for errors
        # and if the client exists in the first place
        if "client" in amp_client_detail['amp:amp_client_detail'] and amp_client_detail['amp:amp_client_detail']['client']['assoc_stat'] == 'true':

            # Make a shoter version of amp_client_detail['amp:amp_client_detail']['client']
            # for easier asignment to users_ap_info dict
            client_ap_bulk = amp_client_detail['amp:amp_client_detail']['client']

            self.users_ap_info['ap_id']  = client_ap_bulk['ap'].get('@id')
            self.users_ap_info['name']   = client_ap_bulk['ap'].get('#text')
            self.users_ap_info['radio']  = client_ap_bulk.get('radio_mode')
            self.users_ap_info['essid']  = client_ap_bulk.get('ssid')
            self.users_ap_info['vlan']   = client_ap_bulk.get('vlan')
            self.users_ap_info['client_aw_url'] = client_detail_resp.request.url

            # Make another call to AW, this time with the ID of the AP to get more info
            # on it including the controllers FQDN and ID
            ap_detail_response = self.session.get(ap_list_url, params={'id': self.users_ap_info['ap_id']}, timeout=30)
            logger.debug(f"Called URL: {ap_detail_response.request.url}")
            logger.debug(f"The response: {ap_detail_response.content}")
            ap_detail_response.raise_for_status()

            # Convert the XML ap_detail_response into a dictionary
            ap_detail = xmltodict.parse(ap_detail_response.content)['amp:amp_ap_list']['ap']
            logger.debug(f"AP detail info:\n{ap_detail}")

            self.users_ap_info['controller_id']  = ap_detail.get('controller_id')
            self.users_ap_info['client_count']   = ap_detail.get('client_count')
            self.users_ap_info['firmware']       = ap_detail.get('firmware')
            self.users_ap_info['ap_fqdn']        = ap_detail.get('fqdn')
            self.users_ap_info['lan_ip']         = ap_detail.get('lan_ip')
            self.users_ap_info['lan_mac']        = ap_detail.get('lan_mac')
            self.users_ap_info['model']          = ap_detail['model'].get('#text')
            self.users_ap_info['operating_mode'] = ap_detail.get('operating_mode')
            self.users_ap_info['serial_number']  = ap_detail.get('serial_number')
            self.users_ap_info['ap_aw_url']      = ap_detail_response.request.url

            logger.info(f"AirWave Client's URL:\n{self.users_ap_info['client_aw_url']}")
            logger.info(f"AirWave client's associated AP:\n{self.users_ap_info['name']}")
            logger.info(f"AirWave client's ESSID: {self.users_ap_info['essid']}")
            logger.info(f"AirWave client's radio: {self.users_ap_info['radio']}")
            logger.info(f"AirWave client's VLAN: {self.users_ap_info['vlan']}")

        elif 'error' in amp_client_detail['amp:amp_client_detail']:
            logger.error(f"This MAC {users_mac} is not valid!")
            logger.error(f"Error output:\n{amp_client_detail['amp:amp_client_detail']['error']}")

        else:
            logger.error(f"Couldn't get data from AirWave for client: {users_mac}")

        logger.debug(f"User's info:\n{self.users_ap_info}")
        return self.users_ap_info

    def get_users_controller_info(self, users_mac=str(), users_controller_id=str()):
        """Should return user's controller info to which the `users_mac` is
        associated with.

        Alternatively, you can pass in the user's controller `@id`.

        Args
        ----
        users_mac: string
            User's MAC address
        users_controller_id: string
            Is a controller's @id.

        Example output
        --------------

        {
            'lan_ip': '10.23.9.12',
            'lan_mac': '20:3f:da:42:aa:76',
            'name': 'wi11',
            'serial_number': '3543543567',
            'device_category': 'controller',
            'controller_id': None,
            'fqdn': 'wi11-loop.blah.com',
            'manufacturer': 'Aruba',
            'model': '7030',
            'is_rap': None
        }
        """
        logger.info('Calling get_users_controller_info()')

        # If you don't pass in the controller Id on which to check on,
        # we need to get it from the passed in users MAC or
        # from a previous event that caused the creation of `self.users_ap_info` variable
        if not users_controller_id:
            if hasattr(self, 'users_ap_info'):
                ap_info = self.users_ap_info
            elif users_mac:
                ap_info = self.get_users_ap_info(users_mac)
            else:
                print("Can't get controller ID if you don't give me users MAC address or it's controller Id!")
                return

            return self.get_all_items_inventory()[ap_info['controller_id']]

        else:
            return self.get_all_items_inventory()[users_controller_id]

    def get_controller_fqdn_list(self):
        """Returns a set of all controllers FQDNs

        Example output
        --------------
        {
            'aa-wi0-loop.blah.com',
            'ab-wi0-loop.blah.com',
            'ac-wi0-loop.blah.com',
            ...
        }
        """
        logger.info('Calling get_controller_fqdn_list()')

        cinv = self.get_controller_inventory()
        cfqdn = set()

        for controller in cinv.values():
            cfqdn.add(controller)

        return cfqdn

    def get_multiple_aps_controllerid(self, ap_names=None):
        """I have no idea what I wante to achieve with this one TBH.
        Returns a list of controllers `@id` or `None` to which the list of
        APs passed in belong to?

        Args
        ----
        `ap_names`: list
            List of AP names
        """
        logger.info('Calling get_multiple_aps_controllerid()')

        ap_db = self.get_apname_to_controllerid_inventory()
        ctrl_ids = set()

        if ap_names and isinstance(ap_names, list):
            { ctrl_ids.add(ap_db[ap]) for ap in ap_names if ap in ap_db }

        # Return AP's controller's `@id`
        return ctrl_ids

    def get_aps_controller(self, ap_name):
        """Returns the APs controller `@id` to which that `ap_name` is
        associated with.

        You can use this response for getting the controllers FQDN for example.

        Args
        ----
        ap_name: str
            The AP name for which to get the associated controller for.

        Returns
        -------
        The APs controlelr @id
        """
        logger.info('Calling get_aps_controller()')

        ap_db = self.get_apname_to_controllerid_inventory()

        if ap_name in ap_db:
            return ap_db[ap_name]

    # METHODS THAT NEED WORK ARE FROM HERE ONWARDS
    def get_iapvcs_aps(self):
        """
        Returns
        -------

        A dict with all VC's FQDNs as keys and list of their APs
        as values.

        Example output
        --------------

        iapvc_aps = {
            {'zz-iapvc0.blah.com': ['zz-iap01']}
        }
        """
        pass

    def get_controllers_aps(self):
        """Not implemented currently, so don't use

        Returns a list of APs belonging to the 'controller' (passed in as FQDN

        DOES NOT return APs that belong to VCs

        Return
        ------
        A dict of controller FQDNs as keys and that controller's list of APs

        Example output
        --------------
        {
            'loop0-cont01.blah.com': ['ap01'],
            'loop0-cont02.blah.com': [
                'ap01',
                'ap09',
                'ap06',
                'ap13',
                'ap05',
                'ap02',
                'ap07',
            ],
            ...
        }
        """
        logger.info('Calling get_controllers_aps()')

        # Get relevant databases
        controller_inventory_db = self.get_controller_inventory()
        #print(controller_inventory_db)
        contid_to_aplist_db = self.get_controllerid_to_ap_inventory()
        #print(contid_to_aplist_db)

        # dict with relevant controllers and their AP list
        cnames_to_aps_db = dict()

        # Write value@key from `controller_inventory_db` as key and value@key from `ap_db`
        # as value into `controller_aps`
        for cid, controller in controller_inventory_db.items():
            if contid_to_aplist_db.get(cid):
                cnames_to_aps_db[controller] = contid_to_aplist_db[cid]

        return cnames_to_aps_db



import sys
import os
import json
import responses
import requests
import unittest
from xmltodict import parse
from mock import patch

sys.path.append(os.path.dirname(__file__) + "/../")
#print(sys.path)

from mmclient import MMClient
from .test_data.mmclient_data import *

BASE_URL = "https://test.arubamm.com"
BASE_API_URL = BASE_URL + ":4343/v1"
LOGIN_URL = BASE_API_URL + "/api/login"

import logging
import logzero
from logzero import logger



class TestMMClient(unittest.TestCase):
    '''Test class for testing MMClient.
    '''
    @responses.activate
    def setUp(self):
        '''MMClient instance creator for the whole class.
        '''
        responses.add(responses.POST, LOGIN_URL, status=200, json=login_resp)
        resp = requests.post(LOGIN_URL)

        # For figuring out how to use responses
        assert resp.json() == login_resp

        self.mmc = MMClient(BASE_URL, "care", "pare")
        self.mmc.comms()

    def test_kwargs_modify_get(self):
        '''Test for kwargs_modify for GET method with endpoint
        'configuration/object/ap_sys_prof'

        Keys must have `endpoint`, `search` and `method = GET` in them
        '''
        expected_kwargs = {
            'endpoint': 'configuration/object/ap_sys_prof',
            'search': 'ap_sys_prof',
            'method': 'GET'}

        actual_kwargs = self.mmc._kwargs_modify('configuration/object/ap_sys_prof')

        self.assertEqual(
            expected_kwargs,
            actual_kwargs,
            f'\nExp kwargs = {expected_kwargs}\nAct kwargs = {actual_kwargs}')

    def test_kwargs_modify_post(self):
        '''Test for kwargs_modify for POST method with endpoint 'configuration/object/ap_sys_prof'

        Keys must have `endpoint`, `jpayload` and `method = POST` in them
        '''
        expected_kwargs = {
            'endpoint': 'configuration/object/ap_sys_prof',
            'jpayload': 'some JSON data payload',
            'method': 'POST'}

        actual_kwargs = self.mmc._kwargs_modify(
            'configuration/object/ap_sys_prof',
            data="some JSON data payload")

        self.assertEqual(
            expected_kwargs,
            actual_kwargs,
            f'\nExp kwargs = {expected_kwargs}\nAct kwargs = {actual_kwargs}')

    def test_params_1(self):
        '''Test for _params() by passing in just
        `profile_name='test-01.ap_sys_prof'` with endpoint
        `configuration/object/ap_sys_prof`
        '''
        expected_params = {
            'config_path': '/md',
            'UIDARUBA': 'fntoken',
            'filter': '[{"ap_sys_prof.profile-name": {"$eq": ["test-01.ap_sys_prof"]}}]'
            }

        modified_kwargs = self.mmc._kwargs_modify('configuration/object/ap_sys_prof')

        actual_params = self.mmc._params(
            profile_name='test-01.ap_sys_prof',
            **modified_kwargs)

        self.assertEqual(
            expected_params,
            actual_params,
            f'\nExp params = {expected_params}\nAct params = {actual_params}')

    def test_params_2(self):
        '''Testing _params() by passing in `profile_name='test-01.ap_sys_prof'` AND `config_path='/md/Test'`
        '''
        expected_params = {
            'config_path': '/md/Test',
            'UIDARUBA': 'fntoken',
            'filter': '[{"ap_sys_prof.profile-name": {"$eq": ["test-01.ap_sys_prof"]}}]'
            }

        modified_kwargs = self.mmc._kwargs_modify('configuration/object/ap_sys_prof')
        actual_params = self.mmc._params(
                profile_name='test-01.ap_sys_prof',
                config_path='/md/Test',
                **modified_kwargs)
        self.assertEqual(
            expected_params,
            actual_params,
            f'\nExp params = {expected_params}\nAct params = {actual_params}')

    def test_params_3(self):
        '''Testing _params() by passing in only `config_path='/md/Test'`
        '''
        expected_params = {
            'config_path': '/md/Test',
            'UIDARUBA': 'fntoken',
            }

        modified_kwargs = self.mmc._kwargs_modify('configuration/object/ap_sys_prof')
        actual_params = self.mmc._params(
                config_path='/md/Test',
                **modified_kwargs)
        self.assertEqual(
            expected_params,
            actual_params,
            f'\nExp params = {expected_params}\nAct params = {actual_params}')

    def test_params_4(self):
        '''Testing _params() by passing in the full filter `[ {"ap_sys_prof.profile-name" : { "$in" : ["a_profile"] } } ]` as JSON or list
        '''
        expected_params = {
            'config_path': '/md/Test',
            'UIDARUBA': 'fntoken',
            'filter': '[{"ap_sys_prof.profile-name": {"$in": ["a_profile"]}}]'
            }

        modified_kwargs = self.mmc._kwargs_modify('configuration/object/ap_sys_prof')
        actual_params = self.mmc._params(
                filter=[{"ap_sys_prof.profile-name":{"$in": ["a_profile"]}}],
                config_path='/md/Test',
                **modified_kwargs)
        self.assertEqual(
            expected_params,
            actual_params,
            f'\nExp params = {expected_params}\nAct params = {actual_params}')


    def test_params_5(self):
        '''Testing _params() by passing in the full filter `[ {"ap_sys_prof.profile-name" : { "$in" : ["a_profile"] } } ]` as string
        '''
        expected_params = {
            'config_path': '/md/Test',
            'UIDARUBA': 'fntoken',
            'filter': '[{"ap_sys_prof.profile-name":{"$in": ["a_profile"]}}]'
            }

        modified_kwargs = self.mmc._kwargs_modify('configuration/object/ap_sys_prof')
        actual_params = self.mmc._params(
                filter='[{"ap_sys_prof.profile-name":{"$in": ["a_profile"]}}]',
                config_path='/md/Test',
                **modified_kwargs)
        self.assertEqual(
            expected_params,
            actual_params,
            f'\nExp params = {expected_params}\nAct params = {actual_params}')


if __name__ == "__main__":
    unittest.main()



from .airwave import AirWave
#from .clearpass import ClearPass, ClearPassDB
from .mmclient import MMClient

from ._version import get_versions
__version__ = get_versions()['version']
del get_versions
